package net.sf.sido.pojo;

import com.google.common.base.Function;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import net.sf.sido.DataObject;
import net.sf.sido.DataType;
import net.sf.sido.array.Array;
import net.sf.sido.array.IndexedArray;
import net.sf.sido.common.AnnotatedSupport;
import org.apache.commons.lang.ObjectUtils;

import java.util.*;

public abstract class DOObject extends AnnotatedSupport {

    public static final String DOOBJECT_ARRAY_SOURCE_HASH = "DOObject.array.sourceHash";
    private final DOSchema schema;
    protected final DataObject o;

    protected DOObject(Class<? extends DOSchema> schemaClass, String typeName) {
        // Gets the DOFactory
        DOFactory doFactory = DOFactory.getContextFactory();
        // Gets the schema
        schema = doFactory.getSchema(schemaClass);
        // Creates the data object
        o = doFactory.getDataFactory().newInstance(typeName);
    }

    public DOObject(DOSchema schema, DataObject o) {
        this.schema = schema;
        this.o = o;
    }

    /**
     * Creates a copy of this object
     * @see net.sf.sido.DataObject#copy()
     */
    public <D extends DOObject> D copy() {
        DataObject c = o.copy();
        return (D) schema.create(c);
    }

    /**
     * Creates an immutable copy of this object
     * @see net.sf.sido.DataObject#immutable()
     */
    public <D extends DOObject> D immutable() {
        DataObject c = o.immutable();
        return (D) schema.create(c);
    }

    public DOSchema getSchema() {
        return schema;
    }

    public <S extends DOSchema> S getSchema(Class<S> schemaClass) {
        return schema.getSchema(schemaClass);
    }

    public DataObject getDataObject() {
        return o;
    }

    public DataType getDataType() {
        return o.getSidoType();
    }

    public static Function<DOObject, DataObject> getDataObjectFn() {
        return new Function<DOObject, DataObject>() {
            @Override
            public DataObject apply(DOObject i) {
                return i.getDataObject();
            }
        };
    }

    public static <D extends DOObject> Function<D, String> getAsString(final String propertyName) {
        return new Function<D, String>() {
            @Override
            public String apply(D input) {
                Object value = input.getDataObject().get(propertyName);
                return ObjectUtils.toString(value, null);
            }
        };
    }

    public static <D extends DOObject, K> Function<? super D, K> getFn(final String propertyName) {
        return new Function<D, K>() {
            @Override
            public K apply(D input) {
                K value = (K) input.getDataObject().get(propertyName);
                return value;
            }
        };
    }

    @Override
    public int hashCode() {
        return o.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof DOObject) {
            return this.o.equals(((DOObject) obj).o);
        } else {
            return false;
        }
    }

    @Override
    public String toString() {
        return o.toString();
    }

    protected <T extends DOObject> Array<T> getPojoArray(Array<T> previous, Iterable<? extends T> newValues, Array<DataObject> source) {
        if (source == null) {
            return null;
        } else {
            // Gets the hash of the source
            int sourceHash = source.hashCode();
            // Gets the previous hash
            if (previous != null) {
                int previousHash = previous.<Integer>getAnnotation(DOOBJECT_ARRAY_SOURCE_HASH);
                if (previousHash == sourceHash) {
                    return previous;
                }
            }
            // Gets the list of values
            Iterable<T> collection = Collections.<T>emptyList();
            if (previous != null) {
                collection = Iterables.concat(collection, previous);
            }
            if (newValues != null) {
                collection = Iterables.concat(collection, newValues);
            }
            // Indexes the pojo-array per ID
            Map<Integer, T> pojos = new HashMap<Integer, T>(Maps.uniqueIndex(collection, DOUtils.GetSidoIdFn));
            // Creates the new list
            List<T> list = new ArrayList<T>();
            // For each sido-item
            for (DataObject o : source) {
                int sidoId = o.getSidoId();
                // Gets the corresponding previous POJO
                T pojo = pojos.get(sidoId);
                if (pojo == null) {
                    pojo = getSchema().<T>create(o);
                }
                // Adds it to the list
                list.add(pojo);
            }
            // Sets the hash
            Array<T> result = new Array<T>(list);
            result.putAnnotation(DOOBJECT_ARRAY_SOURCE_HASH, sourceHash);
            // OK
            return result;
        }
    }

    protected <T extends DOObject> Array<T> getPojoArray(Array<T> previous, Array<DataObject> source) {
        return getPojoArray(previous, null, source);
    }

    protected <T extends DOObject, K> IndexedArray<T, K> getPojoIndexedArray(IndexedArray<T, K> previous, Iterable<T> newValues, IndexedArray<DataObject, K> source, String indexProperty) {
        if (source == null) {
            return null;
        } else {
            // Gets the hash of the source
            int sourceHash = source.hashCode();
            // Gets the previous hash
            if (previous != null) {
                int previousHash = previous.<Integer>getAnnotation(DOOBJECT_ARRAY_SOURCE_HASH);
                if (previousHash == sourceHash) {
                    return previous;
                }
            }
            // Gets the list of values
            Iterable<T> collection = Collections.<T>emptyList();
            if (previous != null) {
                collection = Iterables.concat(collection, previous);
            }
            if (newValues != null) {
                collection = Iterables.concat(collection, newValues);
            }
            // Indexes the pojo-array per ID
            Map<Integer, T> pojos = new HashMap<Integer, T>(Maps.uniqueIndex(collection, DOUtils.GetSidoIdFn));
            // Creates the new list
            List<T> list = new ArrayList<T>();
            // For each sido-item
            for (DataObject o : source) {
                int sidoId = o.getSidoId();
                // Gets the corresponding previous POJO
                T pojo = pojos.get(sidoId);
                if (pojo == null) {
                    pojo = getSchema().<T>create(o);
                }
                // Adds it to the list
                list.add(pojo);
            }
            // Sets the hash
            IndexedArray<T, K> result = new IndexedArray<T, K>(DOObject.<T, K>getFn(indexProperty), list);
            result.putAnnotation(DOOBJECT_ARRAY_SOURCE_HASH, sourceHash);
            // OK
            return result;
        }
    }

    protected <T extends DOObject, K> IndexedArray<T, K> getPojoIndexedArray(IndexedArray<T, K> previous, IndexedArray<DataObject, K> source, String indexProperty) {
        return getPojoIndexedArray(previous, null, source, indexProperty);
    }
}
